/*******************************************************************************
 *
 * PSCWATCH.C - Passive Service Check Watchdog for NetSaint
 * Version: 1.0b3
 * License: GPL
 * Copyright (c) 2000 Ethan Galstad (netsaint@netsaint.org)
 *
 * Last Modified: 11-16-2000
 *
 * Command line: PSCWATCH <config_file>
 *
 * Description:
 *
 * This program is designed to watch the NetSaint status log and schedule
 * immediate checks of services that are older than a certain freshness
 * threshold.  This is critical when relying on passive service checks in a
 * distributed monitoring setup...
 *
 ******************************************************************************/

#include "../common/common.h"
#include "../common/config.h"

#ifdef USE_MYSQL
#include <mysql/mysql.h>
#include <mysql/errmsg.h>
#endif


#define PROGRAM_VERSION "1.0b3"
#define MODIFICATION_DATE "11-17-2000"

#define MYSQL_SERVICESTATUS_TABLE	"servicestatus"

int refresh_rate=120;
int freshness_threshold=960;
int debug=FALSE;
int process_disabled_checks=FALSE;
int process_nonscheduled_checks=FALSE;
int force_checks=FALSE;

char command_file[MAX_INPUT_BUFFER]="";

char status_file[MAX_INPUT_BUFFER]="";

char xsdmysql_host[MAX_INPUT_BUFFER]="";
char xsdmysql_database[MAX_INPUT_BUFFER]="";
char xsdmysql_username[MAX_INPUT_BUFFER]="";
char xsdmysql_password[MAX_INPUT_BUFFER]="";
int xsdmysql_port=3306;

#ifdef USE_MYSQL
MYSQL xsdmysql;
#endif

typedef struct servicestruct{
	char *host_name;
	char *service_description;
	int  freshness_threshold;
	struct servicestruct *next;
        }service;

service *service_list=NULL;
int exclude_service_list=TRUE;


int read_config_file(char *);
int read_service_file(char *);
void sighandler(int);
int schedule_service_check(char *,char *,time_t);
int should_service_be_processed(char *,char *);
int add_service(char *,char *,int);
service *find_service(char *,char *);
int free_memory(void);
void strip(char *);
void sighandler(int);

#ifdef USE_MYSQL
void check_status_table(void);
#endif
#ifndef USE_MYSQL
void check_status_log(void);
#endif




int main(int argc, char **argv){
	int error=FALSE;
	int result;
	char config_file[MAX_INPUT_BUFFER];
	char buffer[MAX_INPUT_BUFFER];

	/* check command line arguments */
	if(argc!=2)
		error=TRUE;

	if(error==TRUE){

		printf("\n");
		printf("PSCWATCH - Passive Service Check Watchdog for NetSaint\n");
		printf("Copyright (c) 2000 Ethan Galstad (netsaint@netsaint.org)\n");
		printf("Version: %s\n",PROGRAM_VERSION);
		printf("Last Modified: %s\n",MODIFICATION_DATE);
		printf("License: GPL\n");
#ifdef USE_MYSQL
		printf("Status Data Source: MySQL Database\n");
#endif
#ifndef USE_MYSQL
		printf("Status Data Source: Standard Status Log\n");
#endif
		printf("\n");
		printf("Usage: %s <config_file>\n",argv[0]);
		printf("\n");
		printf("Notes:\n");
		printf("This program is designed to watch the NetSaint status log and\n");
		printf("schedule immediate checks of services that are older than a\n");
		printf("certain freshness threshold.  This is critical when relying\n");
		printf("on passive service checks in a distributed monitoring setup...\n");
		printf("\n");

		exit(STATE_UNKNOWN);
		}

	/* open a connection to the syslog facility */
        openlog("pscwatch",LOG_PID,LOG_DAEMON); 

	syslog(LOG_NOTICE,"Starting up...");

	/* initialize signal handling */
	signal(SIGTERM,sighandler);
	signal(SIGQUIT,sighandler);

	/* grab the config file */
	strncpy(config_file,argv[1],sizeof(config_file)-1);
	config_file[sizeof(config_file)-1]='\x0';

	/* make sure the config file uses an absolute path */
	if(config_file[0]!='/'){

		/* save the name of the config file */
		strncpy(buffer,config_file,sizeof(buffer));
		buffer[sizeof(buffer)-1]='\x0';

		/* get absolute path of current working directory */
		strcpy(config_file,"");
		getcwd(config_file,sizeof(config_file));

		/* append a forward slash */
		strncat(config_file,"/",sizeof(config_file)-2);
		config_file[sizeof(config_file)-1]='\x0';

		/* append the config file to the path */
		strncat(config_file,buffer,sizeof(config_file)-strlen(config_file)-1);
		config_file[sizeof(config_file)-1]='\x0';
	        }

	/* read the config file */
	result=read_config_file(config_file);	

	/* exit if there are errors... */
	if(result==ERROR){
		syslog(LOG_ERR,"Config file '%s' contained errors, bailing out...",config_file);
		return STATE_CRITICAL;
		}

#ifdef USE_MYSQL
	/* initialize MySQL */
        if(!mysql_init(&xsdmysql)){
                syslog(LOG_ERR,"Error: mysql_init() failed, bailing out...\n");
                return STATE_CRITICAL;
                }
#endif

	/* else daemonize and start working... */
	if(fork()==0){

		if(debug==TRUE){
			if(service_list==NULL)
				syslog(LOG_NOTICE,"All services will be processed.");
			else if(exclude_service_list==TRUE)
				syslog(LOG_NOTICE,"All services expect those specified in the service list file will be processed.");
			else
				syslog(LOG_NOTICE,"Only those services specified in the service list file will be processed.");
	                }

		while(1){

#ifdef USE_MYSQL
			/* scan the MySQL service status table */
			check_status_table();
#endif
#ifndef USE_MYSQL
			/* scan the log file */
			check_status_log();
#endif

			/* sleep for a while */
			sleep(refresh_rate);
		        }
	        }

	/* parent process exits, child watches the status log/table... */
        return STATE_OK;
	}




/* read in the configuration file */
int read_config_file(char *filename){
	FILE *fp;
	char input_buffer[MAX_INPUT_BUFFER];
	char *temp_buffer;
	char *varname;
	char *varvalue;
	int line;


	/* open the config file for reading */
	fp=fopen(filename,"r");

	/* exit if we couldn't open the config file */
	if(fp==NULL)
		return ERROR;

	line=0;
	while(fgets(input_buffer,MAX_INPUT_BUFFER-1,fp)){

		line++;

		/* skip comments and blank lines */
		if(input_buffer[0]=='#')
			continue;
		if(input_buffer[0]=='\x0')
			continue;
		if(input_buffer[0]=='\n')
			continue;

		/* get the variable name */
		varname=strtok(input_buffer,"=");
		if(varname==NULL){
			syslog(LOG_ERR,"No variable name specified in config file '%s' - Line %d\n",filename,line);
			return ERROR;
		        }

		/* get the variable value */
		varvalue=strtok(NULL,"\n");
		if(varvalue==NULL){
			syslog(LOG_ERR,"No variable value specified in config file '%s' - Line %d\n",filename,line);
			return ERROR;
		        }

		if(!strcmp(varname,"status_file")){
			if(strlen(varvalue)>sizeof(status_file)-1){
				syslog(LOG_ERR,"Status file name is too long in config file '%s' - Line %d\n",filename,line);
				return ERROR;
			        }
			strncpy(status_file,varvalue,sizeof(status_file)-1);
			status_file[sizeof(status_file)-1]='\x0';
		        }

		else if(!strcmp(varname,"command_file")){
			if(strlen(varvalue)>sizeof(command_file)-1){
				syslog(LOG_ERR,"Command file name is too long in config file '%s' - Line %d\n",filename,line);
				return ERROR;
			        }
			strncpy(command_file,varvalue,sizeof(command_file)-1);
			command_file[sizeof(command_file)-1]='\x0';
		        }

		else if(!strcmp(varname,"refresh_rate")){
			refresh_rate=atoi(varvalue);
			if(refresh_rate<=0){
				syslog(LOG_ERR,"Invalid value for refresh_rate in config file '%s' - Line %d\n",filename,line);
				return ERROR;
			        }
		        }

		else if(!strcmp(varname,"freshness_threshold")){
			freshness_threshold=atoi(varvalue);
			if(freshness_threshold<=0){
				syslog(LOG_ERR,"Invalid value for freshness_threshold in config file '%s' - Line %d\n",filename,line);
				return ERROR;
			        }
		        }

		else if(!strcmp(varname,"mysql_host")){
			if(strlen(varvalue)>sizeof(command_file)-1){
				syslog(LOG_ERR,"MySQL host name is too long in config file '%s' - Line %d\n",filename,line);
				return ERROR;
			        }
			strncpy(xsdmysql_host,varvalue,sizeof(xsdmysql_host)-1);
			xsdmysql_host[sizeof(xsdmysql_host)-1]='\x0';
		        }

		else if(!strcmp(varname,"mysql_database")){
			if(strlen(varvalue)>sizeof(command_file)-1){
				syslog(LOG_ERR,"MySQL database name is too long in config file '%s' - Line %d\n",filename,line);
				return ERROR;
			        }
			strncpy(xsdmysql_database,varvalue,sizeof(xsdmysql_database)-1);
			xsdmysql_database[sizeof(xsdmysql_database)-1]='\x0';
		        }

		else if(!strcmp(varname,"mysql_username")){
			if(strlen(varvalue)>sizeof(command_file)-1){
				syslog(LOG_ERR,"MySQL username is too long in config file '%s' - Line %d\n",filename,line);
				return ERROR;
			        }
			strncpy(xsdmysql_username,varvalue,sizeof(xsdmysql_username)-1);
			xsdmysql_username[sizeof(xsdmysql_username)-1]='\x0';
		        }

		else if(!strcmp(varname,"mysql_password")){
			if(strlen(varvalue)>sizeof(command_file)-1){
				syslog(LOG_ERR,"MySQL password is too long in config file '%s' - Line %d\n",filename,line);
				return ERROR;
			        }
			strncpy(xsdmysql_password,varvalue,sizeof(xsdmysql_password)-1);
			xsdmysql_password[sizeof(xsdmysql_password)-1]='\x0';
		        }

		else if(!strcmp(varname,"mysql_port")){
			xsdmysql_port=atoi(varvalue);
			if(xsdmysql_port<=0){
				syslog(LOG_ERR,"Invalid value for mysql_port in config file '%s' - Line %d\n",filename,line);
				return ERROR;
			        }
		        }

		else if(!strcmp(varname,"debug"))
			debug=(atoi(varvalue)>0)?TRUE:FALSE;

		else if(!strcmp(varname,"process_disabled_checks"))
			process_disabled_checks=(atoi(varvalue)>0)?TRUE:FALSE;

		else if(!strcmp(varname,"process_nonscheduled_checks"))
			process_nonscheduled_checks=(atoi(varvalue)>0)?TRUE:FALSE;

		else if(!strcmp(varname,"force_checks"))
			force_checks=(atoi(varvalue)>0)?TRUE:FALSE;

		else if(!strcmp(varname,"service_list"))
			read_service_file(varvalue);

		else if(!strcmp(varname,"exclude_service_list"))
			exclude_service_list=(atoi(varvalue)>0)?TRUE:FALSE;

		else{
			syslog(LOG_ERR,"Unknown option specified in config file '%s' - Line %d\n",filename,line);

			return ERROR;
		        }

	        }


	/* close the config file */
	fclose(fp);

	return OK;
	}



#ifndef USE_MYSQL
/* check the status log */
void check_status_log(void){
	char input_buffer[MAX_INPUT_BUFFER];
	char *host_name;
	char *svc_description;
	time_t last_check;
	time_t next_check;
	time_t current_time;
	int checks_enabled;
	int accept_passive_checks;
	char *temp_ptr;
	int total_entries=0;
	FILE *fp;
	service *temp_service;
	int threshold;


	/* get the current time */
	time(&current_time);

	/* open the status log */
	fp=fopen(status_file,"r");
	if(fp==NULL){
		if(debug==TRUE)
			syslog(LOG_ERR,"Could not open status log '%s' for reading",status_file);
		return;
	        }

	while(fgets(input_buffer,sizeof(input_buffer)-1,fp)){

		if(feof(fp))
			break;

		total_entries++;

		/* skip this entry if it isn't a service definition */
		if(!strstr(input_buffer,"] SERVICE;"))
			continue;

		temp_ptr=strtok(input_buffer,";");

		host_name=strtok(NULL,";");
		svc_description=strtok(NULL,";");

		if(should_service_be_processed(host_name,svc_description)==FALSE)
			continue;

		temp_ptr=strtok(NULL,";");
		temp_ptr=strtok(NULL,";");
		temp_ptr=strtok(NULL,";");

		temp_ptr=strtok(NULL,";");
		last_check=(time_t)strtoul(temp_ptr,NULL,0);

		temp_ptr=strtok(NULL,";");
		next_check=(time_t)strtoul(temp_ptr,NULL,0);

		temp_ptr=strtok(NULL,";");

		temp_ptr=strtok(NULL,";");
		checks_enabled=(atoi(temp_ptr)>0)?TRUE:FALSE;

		temp_ptr=strtok(NULL,";");
		accept_passive_checks=(atoi(temp_ptr)>0)?TRUE:FALSE;

		/* skip this check if passive check results are not being accepted */
		if(accept_passive_checks==FALSE)
			continue;

		/* if this check is disabled, skip it */
		if(checks_enabled==FALSE && process_disabled_checks==FALSE)
			continue;

		/* if this check is not being scheduled, skip it */
		if(next_check==(time_t)0 && process_nonscheduled_checks==FALSE)
			continue;

		/* determine what freshness threshold we should use for this service */
		temp_service=find_service(host_name,svc_description);
		if(temp_service==NULL)
			threshold=freshness_threshold;
		else if(temp_service->freshness_threshold<=0)
			threshold=freshness_threshold;
		else
			threshold=temp_service->freshness_threshold;

		/* if we've exceeded the threshold, schedule an immediate check of this service */
		if(current_time-last_check>threshold)
			schedule_service_check(host_name,svc_description,current_time);
	        }

	fclose(fp);

	return;
        }
#endif


#ifdef USE_MYSQL
void check_status_table(void){
        char sql_statement[MAX_INPUT_BUFFER];
        char buffer[MAX_INPUT_BUFFER];
        int result;
        MYSQL_RES *query_result=NULL;
        MYSQL_ROW result_row;
	unsigned long last_check;
	time_t current_time;
	service *temp_service;
	int threshold;

        /* establish a connection to the MySQL server */
        if(!mysql_real_connect(&xsdmysql,xsdmysql_host,xsdmysql_username,xsdmysql_password,xsdmysql_database,xsdmysql_port,NULL,0)){
                syslog(LOG_ERR,"Error: Could not connect to MySQL database '%s' on host '%s' using username '%s' and password '%s'!\n",xsdmysql_database,xsdmysql_host,xsdmysql_username,xsdmysql_password);
                return;
                }

	time(&current_time);

        /* construct the SQL statement */
        snprintf(sql_statement,sizeof(sql_statement)-1,"SELECT UNIX_TIMESTAMP(last_check) AS last_check, host_name, service_description FROM %s WHERE accept_passive_checks='1' %s %s",MYSQL_SERVICESTATUS_TABLE,(process_disabled_checks==FALSE)?"AND checks_enabled='1'":"",(process_nonscheduled_checks==FALSE)?"AND should_be_scheduled='1'":"");
        sql_statement[sizeof(sql_statement)-1]='\x0';

        /* run the query... */
        result=mysql_query(&xsdmysql,sql_statement);

        /* there was an error... */
        if(result){
		if(debug==TRUE)
			syslog(LOG_DEBUG,"Could not execute SQL query: %s",sql_statement);
                return;
	        }

        /* get the results of the query */
        if((query_result=mysql_store_result(&xsdmysql))==NULL){
		if(debug==TRUE)
			syslog(LOG_DEBUG,"Could not get results of SQL query: %s",sql_statement);
                return;
	        }

        /* process all returned rows */
        while((result_row=mysql_fetch_row(query_result))!=NULL){

		if(should_service_be_processed(result_row[1],result_row[2])==FALSE)
			continue;

		last_check=strtoul(result_row[0],NULL,10);

		temp_service=find_service(result_row[1],result_row[2]);
		if(temp_service==NULL)
			threshold=freshness_threshold;
		else if(temp_service->freshness_threshold<=0)
			threshold=freshness_threshold;
		else
			threshold=temp_service->freshness_threshold;

		if((unsigned long)current_time-last_check>threshold)
			schedule_service_check(result_row[1],result_row[2],current_time);
	        }

        /* close the connection to the database server */
        mysql_close(&xsdmysql);

	return;
        }
#endif



/* writes immediate service check command to the NetSaint command file */
int schedule_service_check(char *host_name, char *svc_description,time_t schedule_time){
	FILE *fp;

	/* open the command file for appending */
	fp=fopen(command_file,"a");
	if(fp==NULL){
		if(debug==TRUE)
			syslog(LOG_ERR,"Could not open command file '%s' for appending",command_file);
		return ERROR;
	        }

	fprintf(fp,"[%lu] SCHEDULE_%sSVC_CHECK;%s;%s;%lu\n",(unsigned long)schedule_time,(force_checks==TRUE)?"FORCED_":"",host_name,svc_description,(unsigned long)schedule_time);

	fclose(fp);

	return OK;
        }



/* determines whether or not a service should be processed */
int should_service_be_processed(char *host_name, char *svc_description){
	service *temp_service=NULL;
	int has_been_found=FALSE;

	/* if we don't have a list of services to include or exclude, check it... */
	if(service_list==NULL)
		return TRUE;

	/* otherwise scan the list */
	for(temp_service=service_list;temp_service!=NULL;temp_service=temp_service->next){
		if(!strcmp(temp_service->host_name,host_name) && !strcmp(temp_service->service_description,svc_description)){
			has_been_found=TRUE;
			break;
		        }
	        }

	/* we found the service... */
	if(has_been_found==TRUE){

		/* it should be excluded... */
		if(exclude_service_list==TRUE)
			return FALSE;

		/* it should be included... */
		return TRUE;
	        }

	/* we did not find the in the inclusion list, so don't process it */
	else if(exclude_service_list==FALSE)
		return FALSE;

	return TRUE;
        }



/* adds a service */
int add_service(char *host_name, char *svc_description,int freshness_threshold){
	service *new_service=NULL;

	if(host_name==NULL || svc_description==NULL)
		return ERROR;

	strip(host_name);
	strip(svc_description);

	new_service=(service *)malloc(sizeof(service));
	if(new_service==NULL)
		return ERROR;

	new_service->next=NULL;

	new_service->host_name=(char *)malloc(strlen(host_name)+1);
	if(new_service->host_name==NULL){
		free(new_service);
		return ERROR;
	        }

	strcpy(new_service->host_name,host_name);

	new_service->service_description=(char *)malloc(strlen(svc_description)+1);
	if(new_service->service_description==NULL){
		free(new_service->host_name);
		free(new_service);
		return ERROR;
	        }

	strcpy(new_service->service_description,svc_description);

	new_service->freshness_threshold=freshness_threshold;

	new_service->next=service_list;
	service_list=new_service;

	return OK;
        }


/* finds a service */
service *find_service(char *host_name, char *svc_description){
	service *temp_service=NULL;

	if(host_name==NULL || svc_description==NULL)
		return NULL;

	for(temp_service=service_list;temp_service!=NULL;temp_service=temp_service->next){
		if(!strcmp(host_name,temp_service->host_name) && !strcmp(svc_description,temp_service->service_description))
			return temp_service;
	        }

	return NULL;
        }



/* frees memory */
int free_memory(void){
	service *temp_service=NULL;
	service *next_service=NULL;

	for(temp_service=service_list;temp_service!=NULL;temp_service=next_service){
		next_service=temp_service->next;
		free(temp_service->host_name);
		free(temp_service->service_description);
		free(temp_service);
	        }

	return;
        }




/* read in the service file */
int read_service_file(char *filename){
	FILE *fp;
	char input_buffer[MAX_INPUT_BUFFER];
	char *host_name;
	char *svc_description;
	char *temp_ptr;
	int freshness_threshold;


	/* open the config file for reading */
	fp=fopen(filename,"r");

	/* exit if we couldn't open the config file */
	if(fp==NULL){
		syslog(LOG_ERR,"Could not open service file '%s' for reading...",filename);
		return ERROR;
	        }

	while(fgets(input_buffer,MAX_INPUT_BUFFER-1,fp)){

		/* skip comments and blank lines */
		if(input_buffer[0]=='#')
			continue;
		if(input_buffer[0]=='\x0')
			continue;
		if(input_buffer[0]=='\n')
			continue;

		/* get the host name */
		host_name=strtok(input_buffer,";");
		
		/* get the service description */
		svc_description=strtok(NULL,";");

		/* get the freshness threshold (optional) */
		temp_ptr=strtok(NULL,"\n");
		if(temp_ptr==NULL)
			freshness_threshold=-1;
		else
			freshness_threshold=atoi(temp_ptr);
		
		/* add the service */
		if(add_service(host_name,svc_description,freshness_threshold)==ERROR){
			if(debug==TRUE)
				syslog(LOG_ERR,"Could not add service '%s' on host '%s' to exclusion/inclusion list",svc_description,host_name);
		        }
	        }

	/* close the file */
	fclose(fp);

	return OK;
	}



/* strip newline, carriage return, and tab characters from end of a string */
void strip(char *buffer){
	int lastchar;

	if(buffer==NULL)
		return;

	lastchar=strlen(buffer)-1;

	while((buffer[lastchar]=='\n'||buffer[lastchar]=='\r'||buffer[lastchar]==' '||buffer[lastchar]=='\t') && lastchar>=0){
		buffer[lastchar]='\x0';
		lastchar=strlen(buffer)-1;
	        }

	return;
	}


/* handle signals */
void sighandler(int signum){

	syslog(LOG_NOTICE,"Shutting down...");

	/* free memory */
	free_memory();

#ifdef USE_MYSQL
	mysql_close(&xsdmysql);
#endif

	/* terminate program */
	exit(0);

	return;
        }
